module nscrdll;

private import nscr.charset;

private import std.c.windows.windows;
private import std.string;
private import std.conv;

// NScripterエクスポート関数

struct DIBDATA {
    //NScripter内部のDIBのデータ。常に24bitで、DIBの通常の仕様どおりy座標が大きいほうから入っている。
    int width;
    int height;
    int color;
    byte *bits;
}

extern (C){
    typedef void function (int spnum, out DIBDATA d) GetSpriteDIBFuncPtr;
    typedef void function (int spnum, out int width, out int height, out int cellmax) GetSpriteDataFuncPtr;
}

private GetSpriteDIBFuncPtr p_GetSpriteDIB;
private GetSpriteDataFuncPtr p_GetSpriteData;

private void GetSpriteDIB(int spnum, out DIBDATA d){
    p_GetSpriteDIB(spnum, d);
}

private void GetSpriteData(int spnum, out int width, out int height, out int cellmax){
    p_GetSpriteData(spnum, width, height, cellmax);
}

void init(HMODULE hm){
    p_GetSpriteDIB = cast(GetSpriteDIBFuncPtr)GetProcAddress(hm, "GetSpriteDIB");
    p_GetSpriteData = cast(GetSpriteDataFuncPtr)GetProcAddress(hm, "GetSpriteData");
}

// プラグイン

align(1) struct RGBDATA {
    ubyte b;
    ubyte g;
    ubyte r;
    char[] toString() {
        return std.string.toString(r) ~ ":" ~ std.string.toString(g) ~ ":" ~ std.string.toString(b);
    }
    bool isTransparent(){
        if((this.r & 0xff) != 0xff) return false;
        if((this.g & 0xff) != 0xff) return false;
        if((this.b & 0xff) != 0xff) return false;
        return true;
    }
    bool isOpaque(){
        return !isTransparent();
    }
    ubyte asAlpha(){
        // 計算用に逆にする。高速化のためRのみ採用。
        return cast(ubyte)(0xff - r);
    }
}

/// スプライトを表現する。
class Sprite {
    enum Alpha{
        TRANSPARENT = -1,
        OPAQUE = 0
    }
    
    /// スプライトを構築する。
    /**
        @param spnum スプライト番号
    **/
    this(int spnum){
        this.spnum = spnum;
        GetSpriteDIB(spnum, dib);
        GetSpriteData(spnum, width, height, cell);
    }
    
    /// a指定画像かどうか確認する。
    /**
        @retval true a指定
        @retval false a以外の指定
    **/
    bool hasAlpha(){
        return dib.width == width * 2 * cell;
    }
    
    /// 指定セルについて画素ごとに処理する。
    /**
        @param cell セル番号
        @param processPixcel 画素ごとの処理
    **/
    void processInCellByPixcel(int cell, void delegate(int x, int y, inout RGBDATA pix, ubyte alpha) processPixcel){
        RGBDATA* rgb = cast(RGBDATA*)dib.bits;
        int w = width;
        int h = height;
        int withmask = hasAlpha() ? 2 : 1;
        int ox = cell * w * withmask;
        int scanlen = dib.width;
        if(hasAlpha()){
            for(int y = h - 1; y >= 0; y--){
                for(int x = ox + w - 1; x >= ox; x--){
                    int index = y * scanlen + x;
                    processPixcel(x - ox, h - 1 - y, rgb[index], rgb[index + w].asAlpha());
                }
            }
        }
        else{
            for(int y = h - 1; y >= 0; y--){
                for(int x = ox + w - 1; x >= ox; x--){
                    processPixcel(x - ox, h - 1 - y, rgb[y * scanlen + x], 0xff);
                }
            }
        }
    }
    
    version(DEBUG){
        void dump(){
            RGBDATA* rgb = cast(RGBDATA*)dib.bits;
            char[] buf;
            for(int cell = this.cell - 1; cell >= 0; cell--){
                int ox = cell * width * 2 + width;
                for(int y = height - 1; y >= 0; y--){
                    for(int x = ox + width - 1; x >= ox; x--){
                        buf ~= rgb[y * dib.width + x].toString();
                    }
                    buf ~= "\r\n";
                }
            }
            MessageBoxA(null, cast(char*)(buf ~ "\0"), cast(char*)("\0"), MB_OK);
        }
    }
    
    DIBDATA dib;
    int width;
    int height;
    int cell;
    int spnum;
}

extern(C)
{
    export void sp2tex(ubyte* bits, int width, int height, ubyte* param){
        try{
            // 入力パラメータチェック
            if(param == null || param[0] == 0) throw new Exception("フォントのDLLにパラメータを渡していません。");
            char[] paramstr = fromNScrCharsetz(cast(char*)param);
            char[][] params = std.string.split(paramstr, ",");
            int spnum = toInt(params[0]);
            int cellnum = params.length >= 2 ? toInt(params[1]) : -1;
            
            // 関数ポインタの設定
            char[256] fn;
            GetModuleFileNameA(null, cast(char*)fn, fn.sizeof);
            HMODULE hm = LoadLibraryA(cast(char*)fn);
            init(hm);
            FreeLibrary(hm);
            
            // スプライトチェック
            Sprite sprite = new Sprite(spnum);
            if(width < sprite.width || height < sprite.height) throw new Exception("テクスチャサイズは最低でもセル1枚分必要です。");
            if(cellnum < -1 || sprite.cell <= cellnum) throw new Exception("セル番号が不正です。");
            if(cellnum == -1 && width < sprite.width * sprite.cell) throw new Exception("テクスチャサイズが元画像より小さいためフルコピーできません。");
            
            // 処理
            int i = 0; // テクスチャのバイト位置
            int ox = 0; // テクスチャの書き出し位置
            // 転写
            auto processPixcel = delegate void (int x, int y, inout RGBDATA pix, ubyte alpha){
                if(alpha == 0) return;
                i = (ox + x) * 4 + y * width * 4;
                if(bits[i + 3] == 0){
                    bits[i++] = pix.b;
                    bits[i++] = pix.g;
                    bits[i++] = pix.r;
                    bits[i++] = alpha;
                }
                else{
                    int b = bits[i], g = bits[i + 1], r = bits[i + 2];
                    bits[i++] = cast(ubyte)((pix.b - b) * alpha / 255 + b);
                    bits[i++] = cast(ubyte)((pix.g - g) * alpha / 255 + g);
                    bits[i++] = cast(ubyte)((pix.r - r) * alpha / 255 + r);
                }
            };
            if(cellnum == -1){
                //MessageBoxA(null, toStringz(paramstr), "NG!\0", MB_OK);
                int alphacount;
                // 全セル
                int numOfCells = sprite.cell;
                for(int cell = 0; cell < numOfCells; cell++){
                    sprite.processInCellByPixcel(cell, processPixcel);
                    ox += sprite.width;
                }
            }
            else{
                // 指定セル
                sprite.processInCellByPixcel(cellnum, processPixcel);
            }
        }
        catch(Exception e){
            MessageBoxA(null, toNScrCharsetz(e.toString()), toNScrCharsetz("例外発生！"), MB_OK);
        }
    }
}
